﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO.Ports;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

namespace CoscomV4_Tutorial
{
    /// <summary>
    /// Interaktionslogik für MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {

        System.Timers.Timer FailsafeTimer = new System.Timers.Timer();

        /// <summary>
        /// setting decimal point and group seperator
        /// </summary>
        public static NumberFormatInfo ifo = new NumberFormatInfo();

        /// <summary>
        /// Helper for formating time
        /// </summary>
        /// <param name="i"></param>
        /// <returns></returns>
        public static string FormatTime(int i)
        {
            int seconds = i % 60;
            int minutes = (i / 60) % 60;
            int hours = i / 3600;
            if (hours < 1)
                return minutes.ToString("00") + ":" + seconds.ToString("00");
            return hours.ToString("0") + ":" + minutes.ToString("00") + ":" + seconds.ToString("00");
        }

        /// <summary>
        /// Helper for getting right time unit
        /// </summary>
        /// <param name="i"></param>
        /// <returns></returns>
        public static string GetUnitStringForTime(int i)
        {
            int hours = i / 3600;
            if (hours < 1)
                return "mm:ss";
            return "h:mm:ss";
        }


        public MainWindow()
        {
            InitializeComponent();

            // add all available serial ports to combobox
            foreach (string s in SerialPort.GetPortNames())
                AvailablePorts.Items.Add(s);
            if (AvailablePorts.Items.Count > 0)
                AvailablePorts.SelectedIndex = 0;

            // Init failsafe timer
            FailsafeTimer.AutoReset = true;
            FailsafeTimer.Interval = 500; // reset failsafe all 500 ms (1 second is due time, so this has some buffer)
            FailsafeTimer.Elapsed += FailsafeTimer_Elapsed;

            // init number format parser. For this example we only use "." for decimal points regardless of operating system settings
            ifo.NumberDecimalSeparator = ".";
            ifo.NumberGroupSeparator = "";
        }


        private void FailsafeTimer_Elapsed(object sender, System.Timers.ElapsedEventArgs e)
        { // reset failsafe 
            try
            {
                if (MyService != null)
                {
                    MyService.ResetFailsafe();
                }
            }
            catch (Exception er)
            {
                if (er is XDOP.Core.Exception.XDOPDeviceException) // ignore device error response like "not allowed"
                    return;
                MyDevice.DisposeDevice();
                MyDevice = null;
                MyService = null;
                MessageBox.Show("Error setting failsafe.");
                Disconnect_Click(this, new RoutedEventArgs());

            }
        }

        private void SetSpeed_Click(object sender, RoutedEventArgs e)
        {
            if (MyService != null)
            {
                try
                {
                    // set a new drive speed
                    // Note do not check actual number in order that one can simulate error messages from device side if parameter is out of range.
                    MyService.SetSpeed(double.Parse(TargetSpeed.Text, ifo), 0.2);
                }
                catch (Exception er)
                {
                    MessageBox.Show("Error: " + er.Message);
                }
            }
        }

        private void SetElevation_Click(object sender, RoutedEventArgs e)
        {
            if (MyService != null)
            {
                try
                {
                    // set a new elevation value
                    // Note do not check actual number in order that one can simulate error messages from device side if parameter is out of range.
                    MyService.SetElevation(double.Parse(TargetElevation.Text, ifo));
                }
                catch (Exception er)
                {
                    MessageBox.Show("Error: " + er.Message);
                }
            }
        }

        private void Connect_Click(object sender, RoutedEventArgs e)
        {
            Connect.IsEnabled = false;
            // Search the device in a new thread
            Thread t = new Thread(new ParameterizedThreadStart(ConnectInTread));
            t.Start(AvailablePorts.Text);

        }

        XDOP.schemas_coscom_org.device.MCU6Treadmill_1 MyDevice = null; // device instance
        XDOP.schemas_coscom_org.service.ISBasicControlService MyService = null; // service instance
        private void ConnectInTread(object args)
        {
            try
            {
                // Create a new protocol parameter instance. nothing to change here. Default values are ok
                XDOP.Core.ProtocolOptions op = new XDOP.Core.ProtocolOptions();

                // Create a new serial connection parameter for the specified com port and coscom v4 baud rate 19200
                XDOP.Transport.SerialTransport serialTransport = new XDOP.Transport.SerialTransport(new XDOP.Transport.SerialTransportParameter((string)args, 19200));

                MyDevice = new XDOP.schemas_coscom_org.device.MCU6Treadmill_1();

                // connect to the device
                MyDevice.Connect(serialTransport, op);

                // iterate through all services and save a reference to the direct control service
                foreach (CodeGenerator.IServiceObjectV4 service in MyDevice.Services)
                {
                    if (service.ServiceType.Equals(XDOP.schemas_coscom_org.service.BasicControlService.CodeGeneratorServiceType))
                        MyService = (XDOP.schemas_coscom_org.service.ISBasicControlService)service;
                }

                string deviceType = "";
                string serial = "";
                string version = "";
                MyService.GetDeviceInformation(ref deviceType, ref serial, ref version);
                this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { DeviceType, deviceType });
                this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { Serial, serial });
                this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { Firmware, version });

                // check device information if device type is either MCU5 or MCU6
                if (!deviceType.Contains("MCU5") && !deviceType.Contains("MCU6"))
                {
                    MessageBox.Show("Device found but was no MCU device");
                    MyDevice.DisposeDevice();
                    MyDevice = null;
                    MyService = null;
                    return;
                }

                // get speed range
                double minSpeed = 0;
                double maxSpeed = 0;
                MyService.GetSpeedRange(ref minSpeed, ref maxSpeed);
                this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { SpeedRange, minSpeed.ToString("N1") + " - " + maxSpeed.ToString("N1") + " m/s" });

                // get acceleration/deceleration range
                double minAccel = 0;
                double maxAccel = 0;
                MyService.GetAccelDecelRange(ref minAccel, ref maxAccel);
                this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { AccelRange, minAccel.ToString("N1") + " - " + maxAccel.ToString("N1") + " m/s²" });

                // get elevation range
                double minElevation = 0;
                double maxElevation = 0;
                MyService.GetElevationRange(ref minElevation, ref maxElevation);
                this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { ElevationRange, minElevation.ToString("N1") + " - " + maxElevation.ToString("N1") + " %" });


                // subsribe for all service events and request actual value (query)
                MyService.ActualElevationChanged += MyService_ActualElevationChanged;
                MyService_ActualElevationChanged(MyService.ActualElevation); // get first value
                MyService.ActualSpeedChanged += MyService_ActualSpeedChanged;
                MyService_ActualSpeedChanged(MyService.ActualSpeed); // get first value
                MyService.ControlAllowedChanged += MyService_ControlAllowedChanged;
                MyService_ControlAllowedChanged(MyService.ControlAllowed); // get first value
                MyService.ControlStatusChanged += MyService_ControlStatusChanged;
                MyService_ControlStatusChanged(MyService.ControlStatus); // get first value
                MyService.TargetSpeedChanged += MyService_TargetSpeedChanged;
                MyService_TargetSpeedChanged(MyService.TargetSpeed); // get first value
                MyService.TargetElevationChanged += MyService_TargetElevationChanged;
                MyService_TargetElevationChanged(MyService.TargetElevation); // get first value
                MyService.ActualPowerChanged += MyService_ActualPowerChanged;
                MyService_ActualPowerChanged(MyService.ActualPower); // get first value
                MyService.EnergyConsumptionChanged += MyService_EnergyConsumptionChanged;
                MyService_EnergyConsumptionChanged(MyService.EnergyConsumption); // get first value
                MyService.METChanged += MyService_METChanged;
                MyService_METChanged(MyService.MET); // get first value
                MyService.TimeChanged += MyService_TimeChanged;
                MyService_TimeChanged(MyService.Time); // get first value
                MyService.DistanceChanged += MyService_DistanceChanged;
                MyService_DistanceChanged(MyService.Distance); // get first value
                MyService.HeightChanged += MyService_HeightChanged;
                MyService_HeightChanged(MyService.Height); // get first value
                MyService.HeartRateChanged += MyService_HeartRateChanged;
                MyService_HeartRateChanged(MyService.HeartRate); // get first value
                MyService.RRIntervalChanged += MyService_RRIntervalChanged;
                MyService_RRIntervalChanged(MyService.RRInterval); // get first value
                MyService.ErrorsChanged += MyService_ErrorsChanged;
                MyService_ErrorsChanged(MyService.Errors); // get first value
                MyService.SetEventMask(EventMaskControl.GetCurrentEventMask); // enable all events for simplicity. 



                // set connection state buttons and enable control
                this.Dispatcher.Invoke(new SetControlEnabled(SetControlEnabledImplementation), new object[] { Connect, false });
                this.Dispatcher.Invoke(new SetControlEnabled(SetControlEnabledImplementation), new object[] { Disconnect, true });
                this.Dispatcher.Invoke(new SetControlEnabled(SetControlEnabledImplementation), new object[] { gbControl, true });
                this.Dispatcher.Invoke(new SetControlEnabled(SetControlEnabledImplementation), new object[] { gbPersondata, true });
                this.Dispatcher.Invoke(new SetControlEnabled(SetControlEnabledImplementation), new object[] { gbVariables, true });
            }
            catch (Exception ex)
            {
                // set connection state buttons and disable control
                this.Dispatcher.Invoke(new SetControlEnabled(SetControlEnabledImplementation), new object[] { Connect, true });
                this.Dispatcher.Invoke(new SetControlEnabled(SetControlEnabledImplementation), new object[] { Disconnect, false });
                this.Dispatcher.Invoke(new SetControlEnabled(SetControlEnabledImplementation), new object[] { gbControl, false });
                this.Dispatcher.Invoke(new SetControlEnabled(SetControlEnabledImplementation), new object[] { gbPersondata, false });
                this.Dispatcher.Invoke(new SetControlEnabled(SetControlEnabledImplementation), new object[] { gbVariables, false });
                MessageBox.Show("Error: " + ex.Message);
            }
        }

        private void MyService_ErrorsChanged(string newValue)
        {
            this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { tbErrors, newValue });
        }

        private void MyService_RRIntervalChanged(uint newValue)
        {
            this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { tbRRInterval, newValue.ToString() });
        }

        private void MyService_HeartRateChanged(uint newValue)
        {
            this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { tbHeartRate, newValue.ToString() });
        }

        private void MyService_HeightChanged(double newValue)
        {
            this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { tbHeight, newValue.ToString("N2") });
        }

        private void MyService_DistanceChanged(double newValue)
        {
            this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { tbDistance, newValue.ToString("N1") });
        }

        private void MyService_TimeChanged(int newValue)
        {
            this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { tbTime, FormatTime(newValue) });
            this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { tbTimeUnit, GetUnitStringForTime(newValue) });
        }

        private void MyService_METChanged(double newValue)
        {
            this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { tbMET, newValue.ToString("N1") });
        }

        private void MyService_EnergyConsumptionChanged(double newValue)
        {
            this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { tbEnergyConsumption, newValue.ToString("N1") });
        }

        private void MyService_ActualPowerChanged(int newValue)
        {
            this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { tbActualPower, newValue.ToString() });
        }

        private void MyService_TargetElevationChanged(double newValue)
        {
            this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { vTargetElevation, newValue.ToString("N1") });
        }

        private void MyService_TargetSpeedChanged(double newValue)
        {
            this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { vTargetSpeed, newValue.ToString("N1") });
        }

        private void MyService_ControlStatusChanged(byte newValue)
        {
            this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { tbControlStatus, newValue.ToString() });
        }

        private void Disconnect_Click(object sender, RoutedEventArgs e)
        {
            if (System.Threading.Thread.CurrentThread != Dispatcher.Thread)
            { // if invoke is required
                Dispatcher.Invoke(() => { Disconnect_Click(sender, e); });
                return;
            }
            // Dispose the device and set service instance to null
            if (MyDevice != null)
            {
                MyDevice.DisposeDevice();
                MyDevice = null;
                MyService = null;
            }
            Connect.IsEnabled = true;
            gbControl.IsEnabled = false;
            Disconnect.IsEnabled = false;
        }


        private void MyService_ControlAllowedChanged(byte newValue)
        { // 0 = allowed, 1 = pending, 2 = not allowed
            if (newValue == 0 && !FailsafeTimer.Enabled)
            { // if allowed and failsafe timer not already running, enable it otherwise allowance will be removed after 1 second.
                FailsafeTimer.Start();
                this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { FailsafeActive, "Yes" });
            }
            else
            { // if not allowed, stop failsafe timer
                FailsafeTimer.Stop();
                this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { FailsafeActive, "No" });
            }
            this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { ControlAllowed, newValue == 0 ? "Allowed" : (newValue == 1 ? "pending" : "not allowed") });
        }

        private void MyService_ActualSpeedChanged(double newValue)
        { // show new actual speed (m/s²)
            this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { ActualSpeed, newValue.ToString("N2") });
        }

        private void MyService_ActualElevationChanged(double newValue)
        { // show new actual elevation value (%)
            this.Dispatcher.Invoke(new SetControlText(DelegateImplementation), new object[] { ActualElevation, newValue.ToString("N2") });
        }

        // delegate for setting GUI Textbox
        private delegate void SetControlText(UIElement c, string text);
        private void DelegateImplementation(UIElement c, string text)
        {
            if (c is TextBlock)
                ((TextBlock)c).Text = text;
            if (c is TextBox)
                ((TextBox)c).Text = text;
        }

        // delegate for setting GUI Textbox
        private delegate void SetControlEnabled(Control b, bool isEnabled);
        private void SetControlEnabledImplementation(Control b, bool isEnabled)
        {
            b.IsEnabled = isEnabled;
        }

        private void RequestControl_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                if (MyService != null)
                    MyService.RequestControl(RequestControlMessage.Text); // request control with message written in textbox. 45 Characters max. MCU5 will not be able to display the message and ignore it.
            }
            catch (Exception er)
            {
                MessageBox.Show("Error: " + er.Message);
            }
        }

        private void HoldSpeed_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                if (MyService != null)
                    MyService.HoldSpeed();
            }
            catch (Exception er)
            {
                MessageBox.Show("Error: " + er.Message);
            }
        }

        private void HoldElevation_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                if (MyService != null)
                    MyService.HoldElevation();
            }
            catch (Exception er)
            {
                MessageBox.Show("Error: " + er.Message);
            }
        }

        private void Start_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                if (MyService != null)
                    MyService.Start();
            }
            catch (Exception er)
            {
                MessageBox.Show("Error: " + er.Message);
            }
        }

        private void Stop_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                if (MyService != null)
                    MyService.Stop();
            }
            catch (Exception er)
            {
                MessageBox.Show("Error: " + er.Message);

            }
        }

        private void SetPersonData_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                if (MyService != null)
                {
                    char gender = Gender.SelectedIndex == 0 ? 'M' : 'F';
                    int age = 0;
                    int height = 0;
                    double weight = 0;
                    if (!int.TryParse(Age.Text, out age))
                    {
                        // Note do not check actual number in order that one can simulate error messages from device side if parameter is out of range.
                        MessageBox.Show("Please enter a valid number for person age (1 - 150 years.)");
                        return;
                    }
                    if (!int.TryParse(Height.Text, out height))
                    {
                        // Note do not check actual number in order that one can simulate error messages from device side if parameter is out of range.
                        MessageBox.Show("Please enter a valid number for person height (1 - 300 cm.)");
                        return;
                    }
                    if (!double.TryParse(Weight.Text,NumberStyles.Float, ifo, out weight))
                    {
                        // Note do not check actual number in order that one can simulate error messages from device side if parameter is out of range.
                        MessageBox.Show("Please enter a valid number for person weight (1 - 300 kg.)");
                        return;
                    }


                    MyService.SetPersonData(gender, age, height, weight);

                }
            }
            catch (Exception er)
            {
                MessageBox.Show("Error: " + er.Message);
            }
        }

        private void GetPersonData_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                if (MyService != null)
                {
                    char gender = ' ';
                    int age = 0;
                    int height = 0;
                    double weight = 0;
                    MyService.GetPersonData(ref gender, ref age, ref height, ref weight);
                    this.Dispatcher.Invoke(() =>
                    {
                        Age.Text = age.ToString();
                        Height.Text = height.ToString();
                        Weight.Text = weight.ToString("N2");
                        Gender.SelectedIndex = gender == 'M' ? 0 : 1;



                    });
                }
            }
            catch (Exception er)
            {
                MessageBox.Show("Error: " + er.Message);
            }
        }

        private void ResetCounterValues_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                if (MyService != null)
                    MyService.ResetCounterValues();
            }
            catch (Exception er)
            {
                MessageBox.Show("Error: " + er.Message);
            }
        }

        private void Beep_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                int duration;
                if (!int.TryParse(BeepDuration.Text, out duration))
                {
                    MessageBox.Show("Please enter a valid number for beep duration");
                    return;
                }
                if (duration < 0 && duration > 2550)
                {
                    MessageBox.Show("Please enter a value between 0 and 2550 ms for beep duration.");
                    return;

                }
                if (MyService != null)
                    MyService.Beep((byte)(duration / 10));
            }
            catch (Exception er)
            {
                MessageBox.Show("Error: " + er.Message);

            }
        }

        private void SetEventMask_Click(object sender, RoutedEventArgs e)
        { // show event mask window
            EventMaskControl mask = new EventMaskControl();
            mask.SetService(MyService);
            Window w = new Window();
            w.ResizeMode = ResizeMode.NoResize;
            w.Title = "h/p/cosmos coscom v4 - EventMask";
            w.Content = mask;
            mask.Done += (object o, EventArgs ea) => { w.Close(); };
            w.SizeToContent = SizeToContent.WidthAndHeight;
            w.ShowDialog();
        }

    }
}
